from __future__ import absolute_import

from django.conf import settings
from django.db import models

import rules
from rules.contrib.models import RulesModel


class Book(models.Model):
    isbn = models.CharField(max_length=50, unique=True)
    title = models.CharField(max_length=100)
    author = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)

    def __str__(self):
        return self.title


class TestModel(RulesModel):
    class Meta:
        rules_permissions = {"add": rules.always_true, "view": rules.always_true}

    @classmethod
    def preprocess_rules_permissions(cls, perms):
        perms["custom"] = rules.always_true



from .permissions import add_perm, has_perm, perm_exists, remove_perm, set_perm  # noqa
from .predicates import (  # noqa
    Predicate,
    always_allow,
    always_deny,
    always_false,
    always_true,
    is_active,
    is_authenticated,
    is_group_member,
    is_staff,
    is_superuser,
    predicate,
)
from .rulesets import (  # noqa
    RuleSet,
    add_rule,
    remove_rule,
    rule_exists,
    set_rule,
    test_rule,
)

VERSION = (3, 3, 0, "final", 1)



from os.path import abspath, dirname

BASE_DIR = dirname(dirname(abspath(__file__)))

DEBUG = True

ADMINS = [
    ("test@example.com", "Administrator"),
]

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    },
}

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.messages",
    "django.contrib.sessions",
    "rules",
    "testapp",
]

MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
]

AUTHENTICATION_BACKENDS = [
    "rules.permissions.ObjectPermissionBackend",
    "django.contrib.auth.backends.ModelBackend",
]

CACHE_BACKEND = "locmem://"

SECRET_KEY = "thats-a-secret"

ROOT_URLCONF = "testapp.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "debug": DEBUG,
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]



from django.contrib import admin
from django.urls import re_path

from .views import (
    BookCreateView,
    BookDeleteView,
    BookUpdateErrorView,
    BookUpdateView,
    ViewThatRaises,
    ViewWithPermissionList,
    change_book,
    delete_book,
    view_that_raises,
    view_with_object,
    view_with_permission_list,
)

admin.autodiscover()

urlpatterns = [
    re_path(r"^admin/", admin.site.urls),
    # Function-based views
    re_path(r"^(?P<book_id>\d+)/change/$", change_book, name="change_book"),
    re_path(r"^(?P<book_id>\d+)/delete/$", delete_book, name="delete_book"),
    re_path(r"^(?P<book_id>\d+)/raise/$", view_that_raises, name="view_that_raises"),
    re_path(r"^(?P<book_id>\d+)/object/$", view_with_object, name="view_with_object"),
    re_path(
        r"^(?P<book_id>\d+)/list/$",
        view_with_permission_list,
        name="view_with_permission_list",
    ),
    # Class-based views
    re_path(r"^cbv/create/$", BookCreateView.as_view(), name="cbv.create_book"),
    re_path(
        r"^cbv/(?P<book_id>\d+)/change/$",
        BookUpdateView.as_view(),
        name="cbv.change_book",
    ),
    re_path(
        r"^cbv/(?P<book_id>\d+)/delete/$",
        BookDeleteView.as_view(),
        name="cbv.delete_book",
    ),
    re_path(
        r"^cbv/(?P<book_id>\d+)/raise/$",
        ViewThatRaises.as_view(),
        name="cbv.view_that_raises",
    ),
    re_path(
        r"^cbv/(?P<book_id>\d+)/list/$",
        ViewWithPermissionList.as_view(),
        name="cbv.view_with_permission_list",
    ),
    re_path(
        r"^cbv/(?P<book_id>\d+)/change-error/$",
        BookUpdateErrorView.as_view(),
        name="cbv.change_book_error",
    ),
]



from __future__ import absolute_import

from django.http import HttpResponse
from django.views.generic.edit import CreateView, DeleteView, UpdateView

from rules.contrib.views import (
    LoginRequiredMixin,
    PermissionRequiredMixin,
    objectgetter,
    permission_required,
)

from .models import Book


class BookMixin(object):
    def get_object(self):
        return Book.objects.get(pk=self.kwargs["book_id"])


class BookMixinWithError(object):
    def get_object(self):
        raise AttributeError("get_object")


@permission_required("testapp.change_book", fn=objectgetter(Book, "book_id"))
def change_book(request, book_id):
    return HttpResponse("OK")


class BookCreateView(
    LoginRequiredMixin, PermissionRequiredMixin, BookMixin, CreateView
):
    fields = ["title"]
    template_name = "empty.html"
    permission_required = "testapp.create_book"


class BookUpdateView(
    LoginRequiredMixin, PermissionRequiredMixin, BookMixin, UpdateView
):
    fields = ["title"]
    template_name = "empty.html"
    permission_required = "testapp.change_book"


class BookUpdateErrorView(
    LoginRequiredMixin, PermissionRequiredMixin, BookMixinWithError, UpdateView
):
    fields = ["title"]
    template_name = "empty.html"
    permission_required = "testapp.change_book"


@permission_required("testapp.delete_book", fn=objectgetter(Book, "book_id"))
def delete_book(request, book_id):
    return HttpResponse("OK")


class BookDeleteView(
    LoginRequiredMixin, PermissionRequiredMixin, BookMixin, DeleteView
):
    template_name = "empty.html"
    permission_required = "testapp.delete_book"


@permission_required(
    "testapp.delete_book", fn=objectgetter(Book, "book_id"), raise_exception=True
)
def view_that_raises(request, book_id):
    return HttpResponse("OK")


class ViewThatRaises(
    LoginRequiredMixin, PermissionRequiredMixin, BookMixin, DeleteView
):
    template_name = "empty.html"
    raise_exception = True
    permission_required = "testapp.delete_book"


@permission_required(
    ["testapp.change_book", "testapp.delete_book"], fn=objectgetter(Book, "book_id")
)
def view_with_permission_list(request, book_id):
    return HttpResponse("OK")


class ViewWithPermissionList(
    LoginRequiredMixin, PermissionRequiredMixin, BookMixin, DeleteView
):
    template_name = "empty.html"
    permission_required = ["testapp.change_book", "testapp.delete_book"]


@permission_required("testapp.delete_book", fn=objectgetter(Book, "book_id"))
def view_with_object(request, book_id):
    return HttpResponse("OK")



from django.apps import AppConfig


class RulesConfig(AppConfig):
    name = "rules"
    default = True


class AutodiscoverRulesConfig(RulesConfig):
    default = False

    def ready(self):
        from django.utils.module_loading import autodiscover_modules

        autodiscover_modules("rules")



from __future__ import absolute_import

from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.http import Http404, HttpRequest
from django.test import RequestFactory, TestCase
from django.urls import reverse
from django.utils.encoding import force_str
from django.views.generic import CreateView, View

from testapp.models import Book

import rules  # noqa
from rules.contrib.views import AutoPermissionRequiredMixin, objectgetter

from . import TestData


class FBVDecoratorTests(TestData, TestCase):
    def test_objectgetter(self):
        request = HttpRequest()
        book = Book.objects.get(pk=1)

        self.assertEqual(book, objectgetter(Book)(request, pk=1))
        self.assertEqual(book, objectgetter(Book, attr_name="id")(request, id=1))
        self.assertEqual(book, objectgetter(Book, field_name="id")(request, pk=1))

        with self.assertRaises(ImproperlyConfigured):
            # Raise if no `pk` argument is provided to the view
            self.assertEqual(book, objectgetter(Book)(request, foo=1))

        with self.assertRaises(ImproperlyConfigured):
            # Raise if given invalid model lookup field
            self.assertEqual(book, objectgetter(Book, field_name="foo")(request, pk=1))

        with self.assertRaises(Http404):
            # Raise 404 if no model instance found
            self.assertEqual(book, objectgetter(Book)(request, pk=100000))

    def test_permission_required(self):
        # Adrian can change his book
        self.assertTrue(self.client.login(username="adrian", password="secr3t"))
        response = self.client.get(reverse("change_book", args=(1,)))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(force_str(response.content).strip(), "OK")

        # Martin can change Adrian's book
        self.assertTrue(self.client.login(username="martin", password="secr3t"))
        response = self.client.get(reverse("change_book", args=(1,)))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(force_str(response.content).strip(), "OK")

        # Adrian can delete his book
        self.assertTrue(self.client.login(username="adrian", password="secr3t"))
        response = self.client.get(reverse("delete_book", args=(1,)))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(force_str(response.content).strip(), "OK")

        # Martin can *not* create a book
        # Up to Django v2.1, the response was a redirect to login
        self.assertTrue(self.client.login(username="martin", password="secr3t"))
        response = self.client.get(reverse("cbv.create_book"))
        self.assertIn(response.status_code, [302, 403])

        # Martin can *not* delete Adrian's book and is redirected to login
        self.assertTrue(self.client.login(username="martin", password="secr3t"))
        response = self.client.get(reverse("delete_book", args=(1,)))
        self.assertEqual(response.status_code, 302)

        # Martin can *not* delete Adrian's book and an PermissionDenied is raised
        self.assertTrue(self.client.login(username="martin", password="secr3t"))
        response = self.client.get(reverse("view_that_raises", args=(1,)))
        self.assertEqual(response.status_code, 403)

        # Test views that require a list of permissions

        # Adrian has both permissions
        self.assertTrue(self.client.login(username="adrian", password="secr3t"))
        response = self.client.get(reverse("view_with_permission_list", args=(1,)))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(force_str(response.content).strip(), "OK")

        # Martin does not have delete permission
        self.assertTrue(self.client.login(username="martin", password="secr3t"))
        response = self.client.get(reverse("view_with_permission_list", args=(1,)))
        self.assertEqual(response.status_code, 302)

        # Test views that accept a static object as argument
        # fn is passed to has_perm as-is

        self.assertTrue(self.client.login(username="adrian", password="secr3t"))
        response = self.client.get(reverse("view_with_object", args=(1,)))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(force_str(response.content).strip(), "OK")

        self.assertTrue(self.client.login(username="martin", password="secr3t"))
        response = self.client.get(reverse("view_with_object", args=(1,)))
        self.assertEqual(response.status_code, 302)


class CBVMixinTests(TestData, TestCase):
    def test_get_object_error(self):
        self.assertTrue(self.client.login(username="adrian", password="secr3t"))
        with self.assertRaises(AttributeError):
            self.client.get(reverse("cbv.change_book_error", args=(1,)))

    def test_permission_required_mixin(self):
        # Adrian can change his book
        self.assertTrue(self.client.login(username="adrian", password="secr3t"))
        response = self.client.get(reverse("cbv.change_book", args=(1,)))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(force_str(response.content).strip(), "OK")

        # Martin can change Adrian's book
        self.assertTrue(self.client.login(username="martin", password="secr3t"))
        response = self.client.get(reverse("cbv.change_book", args=(1,)))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(force_str(response.content).strip(), "OK")

        # Adrian can delete his book
        self.assertTrue(self.client.login(username="adrian", password="secr3t"))
        response = self.client.get(reverse("cbv.delete_book", args=(1,)))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(force_str(response.content).strip(), "OK")

        # Martin can *not* delete Adrian's book
        # Up to Django v2.1, the response was a redirect to login
        self.assertTrue(self.client.login(username="martin", password="secr3t"))
        response = self.client.get(reverse("cbv.delete_book", args=(1,)))
        self.assertIn(response.status_code, [302, 403])

        # Martin can *not* delete Adrian's book and an PermissionDenied is raised
        self.assertTrue(self.client.login(username="martin", password="secr3t"))
        response = self.client.get(reverse("cbv.view_that_raises", args=(1,)))
        self.assertEqual(response.status_code, 403)

        # Test views that require a list of permissions

        # Adrian has both permissions
        self.assertTrue(self.client.login(username="adrian", password="secr3t"))
        response = self.client.get(reverse("cbv.view_with_permission_list", args=(1,)))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(force_str(response.content).strip(), "OK")

        # Martin does not have delete permission
        # Up to Django v2.1, the response was a redirect to login
        self.assertTrue(self.client.login(username="martin", password="secr3t"))
        response = self.client.get(reverse("cbv.view_with_permission_list", args=(1,)))
        self.assertIn(response.status_code, [302, 403])


class AutoPermissionRequiredMixinTests(TestCase):
    def setUp(self):
        from testapp.models import TestModel

        self.model = TestModel
        self.req = RequestFactory().get("/")
        self.req.user = AnonymousUser()

    def test_predefined_view_type(self):
        class TestView(AutoPermissionRequiredMixin, CreateView):
            model = self.model
            fields = ()

        self.assertEqual(TestView.as_view()(self.req).status_code, 200)

    def test_custom_view_type(self):
        class CustomViewMixin:
            pass

        class TestView(AutoPermissionRequiredMixin, CustomViewMixin, CreateView):
            model = self.model
            fields = ()
            permission_type_map = [
                (CustomViewMixin, "unknown_perm")
            ] + AutoPermissionRequiredMixin.permission_type_map
            raise_exception = True

        with self.assertRaises(PermissionDenied):
            TestView.as_view()(self.req)

    def test_unknown_view_type(self):
        class TestView(AutoPermissionRequiredMixin, View):
            pass

        with self.assertRaises(ImproperlyConfigured):
            TestView.as_view()(self.req)

    def test_overwrite_perm_type(self):
        class TestView(AutoPermissionRequiredMixin, CreateView):
            model = self.model
            fields = ()
            permission_type = "unknown"
            raise_exception = True

        with self.assertRaises(PermissionDenied):
            TestView.as_view()(self.req)

    def test_disable_perm_checking(self):
        class TestView(AutoPermissionRequiredMixin, CreateView):
            model = self.model
            fields = ()
            permission_type = None

        self.assertEqual(TestView.as_view()(self.req).status_code, 200)

    def test_permission_required_passthrough(self):
        class TestView(AutoPermissionRequiredMixin, CreateView):
            model = self.model
            fields = ()
            permission_required = "testapp.unknown_perm"
            raise_exception = True

        with self.assertRaises(PermissionDenied):
            TestView.as_view()(self.req)



from unittest import TestCase

from rules.permissions import (
    ObjectPermissionBackend,
    add_perm,
    has_perm,
    perm_exists,
    permissions,
    remove_perm,
    set_perm,
)
from rules.predicates import always_false, always_true


class PermissionsTests(TestCase):
    @staticmethod
    def reset_ruleset(ruleset):
        for k in list(ruleset.keys()):
            ruleset.pop(k)

    def setUp(self):
        self.reset_ruleset(permissions)

    def tearDown(self):
        self.reset_ruleset(permissions)

    def test_permissions_ruleset(self):
        add_perm("can_edit_book", always_true)
        assert "can_edit_book" in permissions
        assert perm_exists("can_edit_book")
        assert has_perm("can_edit_book")
        with self.assertRaises(KeyError):
            add_perm("can_edit_book", always_false)
        set_perm("can_edit_book", always_false)
        assert not has_perm("can_edit_book")
        remove_perm("can_edit_book")
        assert not perm_exists("can_edit_book")

    def test_backend(self):
        backend = ObjectPermissionBackend()
        assert backend.authenticate("someuser", "password") is None

        add_perm("can_edit_book", always_true)
        assert "can_edit_book" in permissions
        assert backend.has_perm(None, "can_edit_book")
        assert backend.has_module_perms(None, "can_edit_book")
        with self.assertRaises(KeyError):
            add_perm("can_edit_book", always_true)
        set_perm("can_edit_book", always_false)
        assert not backend.has_perm(None, "can_edit_book")
        remove_perm("can_edit_book")
        assert not perm_exists("can_edit_book")



from django.contrib.auth.models import Group, User

import testapp.rules  # noqa .imported to register rules
from testapp.models import Book

ISBN = "978-1-4302-1936-1"


class TestData:
    @classmethod
    def setUpTestData(cls):
        adrian = User.objects.create_user(
            "adrian", password="secr3t", is_superuser=True, is_staff=True
        )

        martin = User.objects.create_user("martin", password="secr3t", is_staff=True)

        editors = Group.objects.create(name="editors")
        martin.groups.add(editors)

        Book.objects.create(
            isbn=ISBN, title="The Definitive Guide to Django", author=adrian
        )



from unittest import TestCase

from rules.predicates import always_false, always_true
from rules.rulesets import (
    RuleSet,
    add_rule,
    default_rules,
    remove_rule,
    rule_exists,
    set_rule,
    test_rule,
)


class RulesetTests(TestCase):
    @staticmethod
    def reset_ruleset(ruleset):
        for k in list(ruleset.keys()):
            ruleset.pop(k)

    def setUp(self):
        self.reset_ruleset(default_rules)

    def tearDown(self):
        self.reset_ruleset(default_rules)

    def test_shared_ruleset(self):
        add_rule("somerule", always_true)
        assert "somerule" in default_rules
        assert rule_exists("somerule")
        assert test_rule("somerule")
        assert test_rule("somerule")
        with self.assertRaises(KeyError):
            add_rule("somerule", always_false)
        set_rule("somerule", always_false)
        assert not test_rule("somerule")
        remove_rule("somerule")
        assert not rule_exists("somerule")

    def test_ruleset(self):
        ruleset = RuleSet()
        ruleset.add_rule("somerule", always_true)
        assert "somerule" in ruleset
        assert ruleset.rule_exists("somerule")
        assert ruleset.test_rule("somerule")
        with self.assertRaises(KeyError):
            ruleset.add_rule("somerule", always_true)
        ruleset.set_rule("somerule", always_false)
        assert not test_rule("somerule")
        ruleset.remove_rule("somerule")
        assert not ruleset.rule_exists("somerule")



# Generated by Django 2.2.2 on 2019-08-05 08:04

import sys

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Book",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("isbn", models.CharField(max_length=50, unique=True)),
                ("title", models.CharField(max_length=100)),
                (
                    "author",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
    ]

    # TestModel doesn't work under Python 2
    if sys.version_info.major >= 3:
        import rules.contrib.models

        operations += [
            migrations.CreateModel(
                name="TestModel",
                fields=[
                    (
                        "id",
                        models.AutoField(
                            auto_created=True,
                            primary_key=True,
                            serialize=False,
                            verbose_name="ID",
                        ),
                    ),
                ],
                bases=(rules.contrib.models.RulesModelMixin, models.Model),
            ),
        ]



#!/usr/bin/env python
from os.path import dirname, join

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from rules import VERSION


def get_version(version):
    """
    Returns a PEP 386-compliant version number from ``version``.
    """
    assert len(version) == 5
    assert version[3] in ("alpha", "beta", "rc", "final")

    parts = 2 if version[2] == 0 else 3
    main = ".".join(str(x) for x in version[:parts])

    sub = ""
    if version[3] != "final":
        mapping = {"alpha": "a", "beta": "b", "rc": "c"}
        sub = mapping[version[3]] + str(version[4])

    return main + sub


with open(join(dirname(__file__), "README.rst")) as f:
    long_description = f.read()


setup(
    name="rules",
    description="Awesome Django authorization, without the database",
    version=get_version(VERSION),
    long_description=long_description,
    url="http://github.com/dfunckt/django-rules",
    author="Akis Kesoglou",
    author_email="akiskesoglou@gmail.com",
    maintainer="Akis Kesoglou",
    maintainer_email="akiskesoglou@gmail.com",
    license="MIT",
    zip_safe=False,
    packages=[
        "rules",
        "rules.templatetags",
        "rules.contrib",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)



from django.contrib.auth.models import User
from django.test import TestCase

from rules.predicates import (
    is_active,
    is_authenticated,
    is_group_member,
    is_staff,
    is_superuser,
)

from . import TestData


class SwappedUser(object):
    pass


class PredicateTests(TestData, TestCase):
    def test_is_authenticated(self):
        assert is_authenticated(User.objects.get(username="adrian"))
        assert not is_authenticated(SwappedUser())

    def test_is_superuser(self):
        assert is_superuser(User.objects.get(username="adrian"))
        assert not is_superuser(SwappedUser())

    def test_is_staff(self):
        assert is_staff(User.objects.get(username="adrian"))
        assert not is_staff(SwappedUser())

    def test_is_active(self):
        assert is_active(User.objects.get(username="adrian"))
        assert not is_active(SwappedUser())

    def test_is_group_member(self):
        p1 = is_group_member("somegroup")
        assert p1.name == "is_group_member:somegroup"
        assert p1.num_args == 1

        p2 = is_group_member("g1", "g2", "g3", "g4")
        assert p2.name == "is_group_member:g1,g2,g3,..."

        p = is_group_member("editors")
        assert p(User.objects.get(username="martin"))
        assert not p(SwappedUser())

        p = is_group_member("editors", "staff")
        assert not p(User.objects.get(username="martin"))
        assert not p(SwappedUser())



from django.contrib.auth.models import User
from django.template import Context, Template
from django.test import TestCase

from testapp.models import Book

from . import ISBN, TestData


class TemplateTagTests(TestData, TestCase):
    tpl_format = """{{% spaceless %}}
        {{% load rules %}}
        {{% {tag} "{name}" user book as can_update %}}
        {{% if can_update %}}
        OK
        {{% else %}}
        NOT OK
        {{% endif %}}
        {{% endspaceless %}}"""

    def test_rule_tag(self):

        # change_book rule

        tpl = Template(self.tpl_format.format(tag="test_rule", name="change_book"))

        # adrian can change his book as its author
        html = tpl.render(
            Context(
                {
                    "user": User.objects.get(username="adrian"),
                    "book": Book.objects.get(isbn=ISBN),
                }
            )
        )
        self.assertEqual(html, "OK")

        # martin can change adrian's book as an editor
        html = tpl.render(
            Context(
                {
                    "user": User.objects.get(username="martin"),
                    "book": Book.objects.get(isbn=ISBN),
                }
            )
        )
        self.assertEqual(html, "OK")

        # delete_book rule

        tpl = Template(self.tpl_format.format(tag="test_rule", name="delete_book"))

        # adrian can delete his book as its author
        html = tpl.render(
            Context(
                {
                    "user": User.objects.get(username="adrian"),
                    "book": Book.objects.get(isbn=ISBN),
                }
            )
        )
        self.assertEqual(html, "OK")

        # martin can *not* delete adrian's book
        html = tpl.render(
            Context(
                {
                    "user": User.objects.get(username="martin"),
                    "book": Book.objects.get(isbn=ISBN),
                }
            )
        )
        self.assertEqual(html, "NOT OK")

    def test_perm_tag(self):

        # change_book permission

        tpl = Template(
            self.tpl_format.format(tag="has_perm", name="testapp.change_book")
        )

        # adrian can change his book as its author
        html = tpl.render(
            Context(
                {
                    "user": User.objects.get(username="adrian"),
                    "book": Book.objects.get(isbn=ISBN),
                }
            )
        )
        self.assertEqual(html, "OK")

        # martin can change adrian's book as an editor
        html = tpl.render(
            Context(
                {
                    "user": User.objects.get(username="martin"),
                    "book": Book.objects.get(isbn=ISBN),
                }
            )
        )
        self.assertEqual(html, "OK")

        # delete_book permission

        tpl = Template(
            self.tpl_format.format(tag="has_perm", name="testapp.delete_book")
        )

        # adrian can delete his book as its author
        html = tpl.render(
            Context(
                {
                    "user": User.objects.get(username="adrian"),
                    "book": Book.objects.get(isbn=ISBN),
                }
            )
        )
        self.assertEqual(html, "OK")

        # martin can *not* delete adrian's book
        html = tpl.render(
            Context(
                {
                    "user": User.objects.get(username="martin"),
                    "book": Book.objects.get(isbn=ISBN),
                }
            )
        )
        self.assertEqual(html, "NOT OK")
